"""
Sample code for Huiwen Jia and Siqian Shen, "Benders Cut Classification via Support Vector Machines for Solving Two-stage Stochastic Programs", INFORMS Jounal on Optimization, 2020.

@author: hwjia@umich.edu
"""

from gurobipy import *
import random
import xlwt
import numpy as np
import time
import math

    
def tr_bd(Warehouse, setup, capacity, Factory, lostsalecost, Pair, cost , Scenario, probability, Demand, gapv, itelimit, timelimit):

    totaltime_start = time.clock()

    m1 = Model("MP")  
    m1.setParam( 'OutputFlag', False )
    x = m1.addVars(Warehouse, vtype=GRB.BINARY, name="x")
    z = m1.addVars(Scenario,vtype=GRB.CONTINUOUS, name="z")
    obj = LinExpr();
    obj += x.prod(setup);
    obj += z.prod(probability);
    m1.setObjective(obj, GRB.MINIMIZE);
    
    
    mastertime=0
    start = time.clock()
    m1.optimize()
    end = time.clock()
    mastertime=mastertime + (end - start)
    zvalue = m1.getAttr('x', z)
    xvalue = m1.getAttr('x', x)
        
    # Begin iteration
    cutssce = {}  
    zup={}
    for s in Scenario:
        cutssce[s] = 0
        zup[s] = math.inf

    #sheet1row=2
    #sheet1column=3
    lowerbound = m1.objVal
    upperbound = (xvalue.prod(setup) + quicksum(zup[s]*probability[s] for s in Scenario)).getValue()
    dif = upperbound - lowerbound
    subtime=0
    t=1 
    mastertimev=[]
    subtimev=[]
    ratiobindv=[]
    iterationv=[]
    realgapv=[]
    cutnv=[]
    lowerboundv=[]
    gaplimit = gapv[len(gapv)-1]
    while ( dif > gaplimit) :
        print("Iteration %d" %t)
        for s in Scenario: 

            ms = Model("SP")
            ms.setParam( 'OutputFlag', False )
            beta = ms.addVars(Factory, vtype=GRB.CONTINUOUS, name="beta") #lostsale
            y = ms.addVars(Pair, vtype=GRB.CONTINUOUS, name="y")
            obj = LinExpr();
            obj += beta.prod(lostsalecost);
            obj += y.prod(cost);
            ms.setObjective(obj, GRB.MINIMIZE);
            ms.addConstrs( (Demand[s,j]- y.sum('*',j)  <= beta[j] for j in Factory), "lostsale")
            ms.addConstrs( ( y.sum(i,'*') <= capacity[i]*xvalue[i] for i in Warehouse ), "Cap_Cons")
            start = time.clock()
            ms.optimize()
            end = time.clock()
            subtime=subtime + (end - start)
            sobj = ms.objVal
            if sobj > zvalue[s] : 
               cutssce[s] = cutssce[s] + 1
               zup[s] = sobj
               sdual = ms.getAttr('Pi') 
               l=0
               pii = {}
               for j in Factory:
                   pii[j] = sdual[l]
                   l=l+1 
               hi = {}
               for i in Warehouse:
                   hi[i] = sdual[l]
                   l=l+1 
               name = '_'.join(["Cut generated in Iteration",str(t),"Scenario",s]) 
               shortname= '_'.join(["I",str(t),"_",s])           
               m1.addConstr( 
               (z[s] >=   quicksum( capacity[i]*x[i]*hi[i] for i in Warehouse)  + quicksum( Demand[s,j] * ( -  pii[j])  for j in Factory) ) , name) 
        start = time.clock()
        m1.optimize()
        end = time.clock()
        mastertime=mastertime + (end - start)
        lowerbound = m1.objVal
        upperbound = min((xvalue.prod(setup) + quicksum(zup[s]*probability[s] for s in Scenario)).getValue(),upperbound)
        if lowerbound==0:
            lowerbound=0.1
        dif = (upperbound - lowerbound)/lowerbound
        print("UB",upperbound,"     LB",lowerbound,"    dif",dif)
        xvalue = m1.getAttr('x', x)
        zvalue = m1.getAttr('x', z)
        t=t+1
        sla=m1.getAttr("Slack") 
        l=0
        nbind=0
        while l < len(sla) :
            if abs(sla[l])<=0.00000001:
                nbind += 1
            l += 1
        ratiobind = (nbind/len(sla))
        cutnv.append(len(sla))
        mastertimev.append(mastertime)
        subtimev.append(subtime)
        iterationv.append(t-1)
        print("Ratiobind",ratiobind)
        ratiobindv.append(ratiobind)
        realgapv.append( dif )
        lowerboundv.append(lowerbound)
       # if t>itelimit:
       #     break
        if mastertime >timelimit:
            break
    
    return (lowerboundv, 
            realgapv,
            iterationv,
            cutnv,
            ratiobindv,
            mastertimev,
            subtimev
            )
    
    
    
    
    
    



